"""
Installation paths.

Map the .data/ subdirectory names to install paths.
"""

import os.path
import sys
import distutils.dist as dist
import distutils.command.install as install


def get_install_command(name):
    # late binding due to potential monkeypatching
    d = dist.Distribution({"name": name})
    i = install.install(d)
    i.finalize_options()
    return i


def get_install_paths(name):
    """
    Return the (distutils) install paths for the named dist.
    
    A dict with ('purelib', 'platlib', 'headers', 'scripts', 'data') keys.
    """
    paths = {}

    i = get_install_command(name)

    for key in install.SCHEME_KEYS:
        paths[key] = getattr(i, "install_" + key)

    # pip uses a similar path as an alternative to the system's (read-only)
    # include directory:
    if hasattr(sys, "real_prefix"):  # virtualenv
        paths["headers"] = os.path.join(
            sys.prefix, "include", "site", "python" + sys.version[:3], name
        )

    return paths



#!/usr/bin/env python
# Pull arguments out of setup.py by monkeypatching setuptools.
# Run in a directory with setup.py to get a starter pyproject.toml and SConstruct
# Daniel Holth <dholth@fastmail.fm>, 2016

from __future__ import print_function, absolute_import

import runpy
from collections import OrderedDict
import setuptools, distutils.core
import sys, os, codecs, errno
import pytoml

import pkgutil

sconstruct_template = pkgutil.get_data("enscons", "SConstruct.in").decode("utf-8")


def find_src_root(metadata):
    """
    Determine source root from src_root or package_dir key.
    Replace package_dir in metadata with src_root.
    Only understands a single root.
    """
    src_root = ""  # instead of '.'
    if "package_dir" in metadata:
        try:
            src_root = metadata["package_dir"][""]
            del metadata["package_dir"]
        except KeyError:
            raise ValueError('Need package_dir mapping "" to a single directory')
    elif "src_root" in metadata:
        src_root = metadata["src_root"]
    metadata["src_root"] = src_root or ""
    return src_root


def _repr(value):
    """repr() without prefix"""
    return repr(value).strip("u")


def gen_sconstruct(metadata):
    """
    Build a starter SConstruct from pyproject metadata.

    Call find_src_root first to try to find metadata['src_root']
    """
    src_root = metadata["src_root"]

    # Include all Python files in py_modules or packages
    py_source = []
    for module in metadata.get("py_modules", []):
        module_path = os.path.join(src_root, module + ".py")
        py_source.append("[" + _repr(module_path) + "]")
    for package in metadata.get("packages", []):
        package_path = os.path.join(src_root, package.replace(".", os.sep), "*.py")
        py_source.append("Glob(" + _repr(package_path).strip("u") + ")")
    py_source = " + ".join(py_source)

    if not py_source:
        py_source = "[]"
        sys.stderr.write("No Python sources?\n")

    return sconstruct_template.format(py_source=py_source, src_root=_repr(src_root))


def write_no_clobber(filename, contents):
    """
    Write to new file `filename` as utf-8 encoded `contents`.
    Warn if existing file is the same as contents or different than contents.
    """
    try:
        with codecs.open(filename, "r", encoding="utf-8") as existing:
            existing_contents = existing.read()
            if existing_contents != contents:
                sys.stderr.write("Not overwriting existing file %s\n" % filename)
            else:
                sys.stderr.write("Existing file %s is up to date\n" % filename)
    except (OSError, IOError) as e:
        if not e.errno == errno.ENOENT:
            raise
        with codecs.open(filename, "w", encoding="utf-8") as f:
            f.write(contents)
            sys.stderr.write("Wrote %s\n" % filename)


def main():
    def setup_(**kw):
        setup_.arguments = kw

    setuptools.setup = setup_
    distutils.core.setup = setup_

    sys.path[0:0] = "."

    runpy.run_module("setup", run_name="__main__")

    # Convert the keys that enscons uses and a few more that are serializable.
    key_order = [
        "name",
        "version",
        "description",
        "classifiers",
        "keywords",
        "author",
        "author_email",
        "maintainer",
        "maintainer_email",
        "url",
        "license",
        "install_requires",
        "extras_require",
        "tests_require",
        "entry_points",
        "long_description",
        "py_modules",
        "packages",
        "package_dir",
    ]

    ordered_arguments = OrderedDict()
    for key in key_order:
        if key in setup_.arguments:
            ordered_arguments[key] = setup_.arguments.pop(key)

    if isinstance(ordered_arguments.get("keywords"), str):
        ordered_arguments["keywords"] = ordered_arguments["keywords"].split()

    if isinstance(ordered_arguments.get("install_requires"), str):
        ordered_arguments["install_requires"] = [ordered_arguments["install_requires"]]

    if isinstance(ordered_arguments.get("py_modules"), str):
        ordered_arguments["py_modules"] = [ordered_arguments["py_modules"]]

    find_src_root(ordered_arguments)

    if "long_description" in ordered_arguments:
        sys.stderr.write("Consider replacing long_description with description_file\n")

    pyproject = pytoml.dumps(
        OrderedDict(
            [
                ["tool", {"enscons": ordered_arguments}],
                [
                    "build-system",
                    {
                        "requires": ["pytoml>=0.1", "enscons"],
                        "build-backend": "enscons.api",
                    },
                ],
            ]
        ),
        sort_keys=False,
    )

    write_no_clobber("pyproject.toml", pyproject)

    sconstruct = gen_sconstruct(ordered_arguments)
    write_no_clobber("SConstruct", sconstruct)


if __name__ == "__main__":
    main()



# Vendored from https://github.com/pypa/packaging/ until its next release:
#
# Copyright (c) Donald Stufft and individual contributors.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import distutils.util
import platform
import re
import sys
import sysconfig
import warnings

import attr

from packaging import tags


INTERPRETER_SHORT_NAMES = {
    "python": "py",  # Generic.
    "cpython": "cp",
    "pypy": "pp",
    "ironpython": "ip",
    "jython": "jy",
}


_32_BIT_INTERPRETER = sys.maxsize <= 2 ** 32


@attr.s(frozen=True, repr=False)
class Tag(object):
    interpreter = attr.ib(converter=str.lower)
    abi = attr.ib(converter=str.lower)
    platform = attr.ib(converter=str.lower)

    def __str__(self):
        return "{}-{}-{}".format(self.interpreter, self.abi, self.platform)

    def __repr__(self):
        return "<{self} @ {self_id}>".format(self=self, self_id=id(self))


def parse_tag(tag):
    tags = set()
    interpreters, abis, platforms = tag.split("-")
    for interpreter in interpreters.split("."):
        for abi in abis.split("."):
            for platform_ in platforms.split("."):
                tags.add(Tag(interpreter, abi, platform_))
    return frozenset(tags)


def _normalize_string(string):
    return string.replace(".", "_").replace("-", "_")


def _cpython_interpreter(py_version):
    # TODO: Is using py_version_nodot for interpreter version critical?
    return "cp{major}{minor}".format(major=py_version[0], minor=py_version[1])


def _use_malloc():
    with_pymalloc = sysconfig.get_config_var("WITH_PYMALLOC")
    if with_pymalloc is None:
        impl = tags.interpreter_name()
        return impl == 'cp' and sys.version_info < (3, 8)
    return False


# TODO: This code is simpler compared to pep425tags as CPython 2.7 didn't seem
#       to need the fallbacks. Is that acceptable?
def _cpython_abi(py_version):
    soabi = sysconfig.get_config_var("SOABI")
    if soabi:
        options = soabi.split("-", 2)[1]
    else:
        found_options = [str(py_version[0]), str(py_version[1])]
        if sysconfig.get_config_var("Py_DEBUG"):
            found_options.append("d")
        if _use_malloc():
            found_options.append("m")
        if sysconfig.get_config_var("Py_UNICODE_SIZE") == 4:
            found_options.append("u")
        options = "".join(found_options)
    return "cp{options}".format(options=options)


def _cpython_tags(py_version, interpreter, abi, platforms):
    for tag in (Tag(interpreter, abi, platform) for platform in platforms):
        yield tag
    for tag in (Tag(interpreter, "abi3", platform) for platform in platforms):
        yield tag
    for tag in (Tag(interpreter, "none", platform) for platform in platforms):
        yield tag
    # PEP 384 was first implemented in Python 3.2.
    for minor_version in range(py_version[1] - 1, 1, -1):
        for platform_ in platforms:
            interpreter = "cp{major}{minor}".format(
                major=py_version[0], minor=minor_version
            )
            yield Tag(interpreter, "abi3", platform_)


def _pypy_interpreter():
    return "pp{py_major}{pypy_major}{pypy_minor}".format(
        py_major=sys.version_info[0],
        pypy_major=sys.pypy_version_info.major,
        pypy_minor=sys.pypy_version_info.minor,
    )


def _generic_abi():
    abi = sysconfig.get_config_var("SOABI")
    if abi:
        return _normalize_string(abi)
    else:
        return "none"


def _pypy_tags(py_version, interpreter, abi, platforms):
    for tag in (Tag(interpreter, abi, platform) for platform in platforms):
        yield tag
    for tag in (Tag(interpreter, "none", platform) for platform in platforms):
        yield tag


def _generic_tags(interpreter, py_version, abi, platforms):
    for tag in (Tag(interpreter, abi, platform) for platform in platforms):
        yield tag
    if abi != "none":
        tags = (Tag(interpreter, "none", platform_) for platform_ in platforms)
        for tag in tags:
            yield tag


def _py_interpreter_range(py_version):
    """
    Yield Python versions in descending order.

    After the latest version, the major-only version will be yielded, and then
    all following versions up to 'end'.
    """
    yield "py{major}{minor}".format(major=py_version[0], minor=py_version[1])
    yield "py{major}".format(major=py_version[0])
    for minor in range(py_version[1] - 1, -1, -1):
        yield "py{major}{minor}".format(major=py_version[0], minor=minor)


def _independent_tags(interpreter, py_version, platforms):
    """
    Return the sequence of tags that are consistent across implementations.

    The tags consist of:
    - py*-none-<platform>
    - <interpreter>-none-any
    - py*-none-any
    """
    for version in _py_interpreter_range(py_version):
        for platform_ in platforms:
            yield Tag(version, "none", platform_)
    yield Tag(interpreter, "none", "any")
    for version in _py_interpreter_range(py_version):
        yield Tag(version, "none", "any")


def _mac_arch(arch, is_32bit=_32_BIT_INTERPRETER):
    if is_32bit:
        if arch.startswith("ppc"):
            return "ppc"
        else:
            return "i386"
    else:
        return arch


def _mac_binary_formats(version, cpu_arch):
    formats = [cpu_arch]
    if cpu_arch == "x86_64":
        if version >= (10, 4):
            formats.extend(["intel", "fat64", "fat32"])
        else:
            return []
    elif cpu_arch == "i386":
        if version >= (10, 4):
            formats.extend(["intel", "fat32", "fat"])
        else:
            return []
    elif cpu_arch == "ppc64":
        # TODO: Need to care about 32-bit PPC for ppc64 through 10.2?
        if version > (10, 5) or version < (10, 4):
            return []
        else:
            formats.append("fat64")
    elif cpu_arch == "ppc":
        if version <= (10, 6):
            formats.extend(["fat32", "fat"])
        else:
            return []

    formats.append("universal")
    return formats


def _mac_platforms(version=None, arch=None):
    version_str, _, cpu_arch = platform.mac_ver()
    if version is None:
        version = tuple(map(int, version_str.split(".")[:2]))
    if arch is None:
        arch = _mac_arch(cpu_arch)
    platforms = []
    for minor_version in range(version[1], -1, -1):
        compat_version = version[0], minor_version
        binary_formats = _mac_binary_formats(compat_version, arch)
        for binary_format in binary_formats:
            platforms.append(
                "macosx_{major}_{minor}_{binary_format}".format(
                    major=compat_version[0],
                    minor=compat_version[1],
                    binary_format=binary_format,
                )
            )
    return platforms


# From PEP 513.
def _is_manylinux_compatible(name, glibc_version):
    # Check for presence of _manylinux module.
    try:
        import _manylinux

        return bool(getattr(_manylinux, name + "_compatible"))
    except (ImportError, AttributeError):
        # Fall through to heuristic check below.
        pass

    return _have_compatible_glibc(*glibc_version)


def _glibc_version_string():
    # Returns glibc version string, or None if not using glibc.
    import ctypes

    # ctypes.CDLL(None) internally calls dlopen(NULL), and as the dlopen
    # manpage says, "If filename is NULL, then the returned handle is for the
    # main program". This way we can let the linker do the work to figure out
    # which libc our process is actually using.
    process_namespace = ctypes.CDLL(None)
    try:
        gnu_get_libc_version = process_namespace.gnu_get_libc_version
    except AttributeError:
        # Symbol doesn't exist -> therefore, we are not linked to
        # glibc.
        return None

    # Call gnu_get_libc_version, which returns a string like "2.5"
    gnu_get_libc_version.restype = ctypes.c_char_p
    version_str = gnu_get_libc_version()
    # py2 / py3 compatibility:
    if not isinstance(version_str, str):
        version_str = version_str.decode("ascii")

    return version_str


# Separated out from have_compatible_glibc for easier unit testing.
def _check_glibc_version(version_str, required_major, minimum_minor):
    # Parse string and check against requested version.
    #
    # We use a regexp instead of str.split because we want to discard any
    # random junk that might come after the minor version -- this might happen
    # in patched/forked versions of glibc (e.g. Linaro's version of glibc
    # uses version strings like "2.20-2014.11"). See gh-3588.
    m = re.match(r"(?P<major>[0-9]+)\.(?P<minor>[0-9]+)", version_str)
    if not m:
        warnings.warn(
            "Expected glibc version with 2 components major.minor,"
            " got: %s" % version_str,
            RuntimeWarning,
        )
        return False
    return (
        int(m.group("major")) == required_major
        and int(m.group("minor")) >= minimum_minor
    )


def _have_compatible_glibc(required_major, minimum_minor):
    version_str = _glibc_version_string()
    if version_str is None:
        return False
    return _check_glibc_version(version_str, required_major, minimum_minor)


def _linux_platforms(is_32bit=_32_BIT_INTERPRETER):
    linux = _normalize_string(distutils.util.get_platform())
    if linux == "linux_x86_64" and is_32bit:
        linux = "linux_i686"
    # manylinux1: CentOS 5 w/ glibc 2.5.
    # manylinux2010: CentOS 6 w/ glibc 2.12.
    manylinux_support = ("manylinux2010", (2, 12)), ("manylinux1", (2, 5))
    manylinux_support_iter = iter(manylinux_support)
    for name, glibc_version in manylinux_support_iter:
        if _is_manylinux_compatible(name, glibc_version):
            platforms = [linux.replace("linux", name)]
            break
    else:
        platforms = []
    # Support for a later manylinux implies support for an earlier version.
    platforms += [linux.replace("linux", name) for name, _ in manylinux_support_iter]
    platforms.append(linux)
    return platforms


def _generic_platforms():
    platform = _normalize_string(distutils.util.get_platform())
    return [platform]


def _interpreter_name():
    name = platform.python_implementation().lower()
    return INTERPRETER_SHORT_NAMES.get(name) or name


def _generic_interpreter(name, py_version):
    version = sysconfig.get_config_var("py_version_nodot")
    if not version:
        version = "".join(map(str, py_version[:2]))
    return "{name}{version}".format(name=name, version=version)


def sys_tags():
    """
    Returns the sequence of tag triples for the running interpreter.

    The order of the sequence corresponds to priority order for the
    interpreter, from most to least important.
    """
    py_version = sys.version_info[:2]
    interpreter_name = _interpreter_name()
    if platform.system() == "Darwin":
        platforms = _mac_platforms()
    elif platform.system() == "Linux":
        platforms = _linux_platforms()
    else:
        platforms = _generic_platforms()

    if interpreter_name == "cp":
        interpreter = _cpython_interpreter(py_version)
        abi = _cpython_abi(py_version)
        for tag in _cpython_tags(py_version, interpreter, abi, platforms):
            yield tag
    elif interpreter_name == "pp":
        interpreter = _pypy_interpreter()
        abi = _generic_abi()
        for tag in _pypy_tags(py_version, interpreter, abi, platforms):
            yield tag
    else:
        interpreter = _generic_interpreter(interpreter_name, py_version)
        abi = _generic_abi()
        for tag in _generic_tags(interpreter, py_version, abi, platforms):
            yield tag
    for tag in _independent_tags(interpreter, py_version, platforms):
        yield tag



"""pytar

Tool-specific initialization for tar (Python stdlib implementation)

Based on SCons' tar and zip builders.

This version supports anonymous tar and a path prefix inside the archive.
"""

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os.path
import tarfile

import SCons.Builder
import SCons.Defaults
import SCons.Node.FS
import SCons.Util

# Do we have to set certain headers for it to "really" be PAX format?
_tarformat = tarfile.PAX_FORMAT


def tar(target, source, env):
    compression = env.get("TARCOMPRESSION", "")
    tarformat = env.get("TARFORMAT", None)
    tarroot = str(env.get("TARROOT", ""))
    tarprefix = str(env.get("TARPREFIX", ""))
    taruid = env.get("TARUID")
    targid = env.get("TARGID")
    tarmtime = env.get("TARMTIME")
    with tarfile.TarFile.open(
        str(target[0]), "w" + (compression and ":") + compression, format=tarformat
    ) as tar:

        def _filter(info):
            """Return potentially anonymize tarinfo"""
            if taruid != None:
                info.uid = taruid
                info.uname = ""
            if targid != None:
                info.gid = targid
                info.gname = ""
            if tarmtime != None:
                info.mtime = tarmtime
            return info

        for s in source:
            arcname = os.path.relpath(str(s), tarroot)
            arcname = os.path.join(tarprefix, arcname)
            tar.add(str(s), arcname=arcname, filter=_filter)


TarAction = SCons.Action.Action(tar, varlist=["TARCOMPRESSION"])

TarBuilder = SCons.Builder.Builder(
    action=SCons.Action.Action("$TARCOM", "$TARCOMSTR"),
    source_factory=SCons.Node.FS.Entry,
    source_scanner=SCons.Defaults.DirScanner,
    suffix="$TARSUFFIX",
    multi=True,
    ensure_suffix=True,
)


def generate(env):
    """Add Builders and construction variables for TAR to an Environment."""
    try:
        bld = env["BUILDERS"]["PyTar"]
    except KeyError:
        bld = TarBuilder
        env["BUILDERS"]["PyTar"] = bld

    env["TAR"] = "tar"
    env["TARFLAGS"] = SCons.Util.CLVar("")
    env["TARCOM"] = TarAction
    env["TARFORMAT"] = _tarformat
    env["TARCOMPRESSION"] = "gz"
    env["TARSUFFIX"] = ".tar"
    # prefix in the internal directory structure
    env["TARPREFIX"] = ""
    env["TARROOT"] = SCons.Util.CLVar("")
    # environment variables to support anonymized tar
    env["TARUID"] = None  # None = use real uid
    env["TARGID"] = None  # None = use real gid
    env["TARMTIME"] = None  # None = use real mtime


def exists(env):
    """Standard library tar function should always exist."""
    return True
